@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2006 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1993-2008 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -17,7 +17,7 @@
  */
 
 /* block.cc : block operations */
-/* $Id: block.cc,v 1.28 2006/04/12 05:46:04 lav Exp $ */
+/* $Id: block.cc,v 1.30 2008/01/21 14:27:40 lav Exp $ */
 
 #include <config.h>
 #include <assert.h>
@@ -782,14 +782,13 @@ next:
       h=TRUE;
 
    if(!h && !View)
-      Message("Block: C-Copy M-Move D-Delete W-Write H-Hide I-Indent U-Unindent R-Read BETXLPA");
+      MessageSync("Block: C-Copy M-Move D-Delete W-Write H-Hide I-Indent U-Unindent R-Read BETXLPA");
    else if(h && !View)
-      Message("Block: B-Begin E-End T-Type L-to Lower P-to uPper X-exchange H-display A-mark All");
+      MessageSync("Block: B-Begin E-End T-Type L-to Lower P-to uPper X-exchange H-display A-mark All");
    else if(h && View)
-      Message("Block: B-Begin E-End T-Type H-display A-mark All");
+      MessageSync("Block: B-Begin E-End T-Type H-display A-mark All");
    else
-      Message("Block: W-Write B-Begin E-End T-Type H-Hide A-mark All");
-   SyncTextWin();
+      MessageSync("Block: W-Write B-Begin E-End T-Type H-Hide A-mark All");
    SetCursor();
    action=GetNextAction();
    flag=REDISPLAY_ALL;
