@@ -16,7 +16,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: highli.cc,v 1.19 2006/11/29 08:14:47 lav Exp $ */
+/* $Id: highli.cc,v 1.20 2007/08/02 08:13:10 lav Exp $ */
 
 #include <config.h>
 
@@ -187,6 +187,7 @@ void InitHighlight()
    unsigned len;
    int res;
    int color,mask;
+   const char *bn=le_basename(FileName);
 
    for(;;)
    {
@@ -251,7 +252,7 @@ void InitHighlight()
 	       }
 	       break;
 	    }
-	    if(fnmatch(s,FileName,0)==0)
+	    if(fnmatch(s,bn,0)==0)
 	    {
 	       match=1;
 	       break;
