@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2000 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1993-2006 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: screen.cc,v 1.29 2000/09/04 08:19:02 lav Exp $ */
+/* $Id: screen.cc,v 1.47 2006/11/29 07:36:35 lav Exp $ */
 
 #include <config.h>
 #include <string.h>
@@ -29,6 +29,7 @@
 #include "block.h"
 #include "highli.h"
 #include "getch.h"
+#include "mb.h"
 
 #ifndef max
 # define max(a,b) ((a)>(b)?(a):(b))
@@ -51,7 +52,8 @@ int    message_sp=0; // number of messages on bottom of screen
 int   range_begin;
 int   range_end;
 
-#if defined(__NCURSES_H) && !defined(mvaddchnstr)
+#if defined(__NCURSES_H) && (!defined(mvaddchnstr) || defined(mvadd_wchnstr))
+#undef mvaddchnstr
 #define mvaddchnstr le_mvaddchnstr
 
 int le_mvaddchnstr(int line, int col, chtype *s, int n)
@@ -186,8 +188,12 @@ void  SyncTextWin()
 
    if(CheckPending()>0)
    {
-      if(++skipped<5)
+      if(++skipped<2000)
       {
+	 attrset(STATUS_LINE_ATTR->n_attr);
+	 move(StatusLineY,COLS-3);
+	 addch(' ');
+	 addch("-\\|/"[time(0)%4]);
 	 leaveok(stdscr,TRUE);
 	 flag=REDISPLAY_ALL;
 	 return;
@@ -206,6 +212,7 @@ void  SyncTextWin()
    if(lim>=line)
       Redisplay(line,ptr,lim);
    flag=0;
+   StatusLine();
 }
 
 int   ScrollBarPos=0;
@@ -228,7 +235,14 @@ void  ScrollBar(int check)
       return;
 
    for(i=TextWinY; i<TextWinY+TextWinHeight; i++)
-      mvaddch(i,ScrollBarX,ACS_CKBOARD);
+   {
+#if USE_MULTIBYTE_CHARS
+      if(mb_mode)
+	 mvadd_wch(i,ScrollBarX,WACS_CKBOARD);
+      else
+#endif
+	 mvaddch(i,ScrollBarX,ACS_CKBOARD);
+   }
    ScrollBarPos=NewPos;
    mvaddch(NewPos+TextWinY,ScrollBarX,' '|A_REVERSE);
 }
@@ -300,9 +314,11 @@ void  ClearMessage()
 void  StatusLine()
 {
    char  status[512];
+   char  status_right[512];
    char  *bn;
    char  name[20];
-   char  chr[4];
+   wchar_t *wname=0;
+   char  chr[8];
    int   l;
    char  flags[16];
 
@@ -312,13 +328,18 @@ void  StatusLine()
       return;
 
    if(Eof())
-      strcpy(chr,"EOF");
+      strcpy(chr,"Ch:EOF");
    else
    {
       if(Eol())
-	 strcpy(chr,"EOL");
+	 strcpy(chr,"Ch:EOL");
       else
-	 sprintf(chr,"%-3d",Char());
+      {
+	 if(MBCheckRight() && MBCharSize>1)
+	    sprintf(chr,"UC:%04X",WChar());
+	 else
+	    sprintf(chr,"Ch:%-3d",Char());
+      }
    }
 
    if(View)
@@ -352,14 +373,31 @@ void  StatusLine()
    {
       bn=le_basename(FileName);
       l=strlen(bn);
+#if USE_MULTIBYTE_CHARS
+      wname=(wchar_t*)alloca((l+1)*sizeof(wchar_t));
+      memset(wname,0,(l+1)*sizeof(wchar_t));
+      mbstowcs(wname,bn,l);
+      l=wcslen(wname);
       if(l>14)
-         sprintf(name,"\"%.*s..%.*s\"",6,bn,6,bn+l-6);
-      else
-         sprintf(name,"\"%s\"",bn);
+      {
+	 memmove(wname+8,wname+l-6,7*sizeof(wchar_t));
+	 wname[6]=wname[7]=L'.';
+      }
+      if(l<=0)
+	 wname=0;
+#endif // USE_MULTIBYTE_CHARS
+      if(wname==0)
+      {
+	 if(l>14)
+	    sprintf(name,"%.*s..%.*s",6,bn,6,bn+l-6);
+	 else
+	    strcpy(name,bn);
+      }
    }
    else
       sprintf(name,"NewFile");
 
+
    if(hex)
       sprintf(status,"OctOffs:0%-11lo",(unsigned long)(Offset()));
    else
@@ -367,20 +405,72 @@ void  StatusLine()
 	 (unsigned long)(GetLine()+1),
 	 (unsigned long)(((Text&&Eol())?stdcol:GetCol())+1));
 
-   sprintf(status+strlen(status),
-      " Size:%-6lu Ch:%3s %s %s Offs:%lu (%d%%)",
-         (unsigned long)(Size()),chr,flags,name,(unsigned long)(Offset()),
-         (int)(Size()?(Offset()*100.+Size()/2)/Size():100));
+   sprintf(status+strlen(status)," Sz:%-6lu %-7s %s",
+         (unsigned long)(Size()),chr,flags);
 
-   l=strlen(status);
-   if(l<COLS)
-      memset(status+l,' ',COLS-l);
-   status[COLS]=0;
+   sprintf(status_right," Offs:%lu (%d%%)",(unsigned long)(Offset()),
+         (int)(Size()?(Offset()*100.+Size()/2)/Size():100));
 
    move(StatusLineY,0);
    SetAttr(STATUS_LINE_ATTR);
+   int prev_x=0;
+   int x=0,y;
    for(bn=status; *bn; bn++)
-      addch_visual((byte)*bn);
+   {
+      prev_x=x;
+      addch((byte)*bn);
+      getyx(stdscr,y,x);
+      if(prev_x>x || (prev_x==x && x==COLS-1))
+	 return;
+   }
+
+   if(x>=COLS-3)
+      return;
+   addch('"');
+   getyx(stdscr,y,x);
+#if USE_MULTIBYTE_CHARS
+   if(wname)
+   {
+      for(wchar_t *w=wname; *w; w++)
+      {
+	 prev_x=x;
+	 wchar_t wc=visualize_wchar(*w);
+	 if(wc!=*w)
+	    attrset(curr_attr->so_attr);
+	 addnwstr(&wc,1);
+	 attrset(curr_attr->n_attr);
+	 getyx(stdscr,y,x);
+	 if(prev_x>x || (prev_x==x && x==COLS-1))
+	    return;
+      }
+   }
+   else // note the following block
+#endif
+   {
+      for(bn=name; *bn; bn++)
+      {
+	 prev_x=x;
+	 addch((byte)*bn);
+	 getyx(stdscr,y,x);
+	 if(prev_x>x || (prev_x==x && x==COLS-1))
+	    return;
+      }
+   }
+   prev_x=x;
+   addch('"');
+   getyx(stdscr,y,x);
+   if(prev_x>x || (prev_x==x && x==COLS-1))
+      return;
+   for(bn=status_right; *bn; bn++)
+   {
+      prev_x=x;
+      addch((byte)*bn);
+      getyx(stdscr,y,x);
+      if(prev_x>x || (prev_x==x && x==COLS-1))
+	 return;
+   }
+   for(int i=COLS-x; i>0; i--)
+      addch(' ');
 }
 
 static unsigned mkhash(byte *data,int len)
@@ -391,10 +481,10 @@ static unsigned mkhash(byte *data,int len)
    return res;
 }
 
-static attr *norm_attr,*blk_attr,*syntax[3];
-static inline attr *FindPosAttr(offs ptr,num line,num col,byte *hlp)
+static const attr *norm_attr,*blk_attr,*syntax[3];
+static const attr *FindPosAttr(offs ptr,num line,num col,byte *hlp)
 {
-   if(col==TextWinWidth-1 && !EolAt(ptr) && !EolAt(ptr+1))
+   if(col>=TextWinWidth-MBCharWidth && !EolAt(ptr) && !EolAt(ptr+MBCharSize))
       return SHADOW_ATTR;
    else if(InBlock(ptr,line+ScrLine,col+ScrShift))
       return blk_attr;
@@ -422,9 +512,15 @@ void  Redisplay(num line,offs ptr,num limit)
    syntax[2]=find_attr(SYNTAX3);
 
    if(!hex)
-      ScreenTop=LineBegin(ScreenTop);
+   {
+      if(!BolAt(ScreenTop))
+	 ScreenTop=TextPoint(ScreenTop.Line(),0);
+   }
    else
-      ScreenTop=ScreenTop&~15;
+   {
+      if(ScreenTop&15)
+	 ScreenTop=ScreenTop&~15;
+   }
 
    if(line<0)
    {
@@ -447,9 +543,15 @@ void  Redisplay(num line,offs ptr,num limit)
    if(ptr<0)
       ptr=0;
 
-   chtype *cl=(chtype*)alloca(max(TextWinWidth,80)*sizeof(chtype));
+   int ll=max(TextWinWidth,80);
+   chtype *cl=(chtype*)alloca(ll*sizeof(chtype));
    chtype *clp;
-   struct attr *ca=norm_attr;
+   const attr *ca=norm_attr;
+
+#ifdef USE_MULTIBYTE_CHARS
+   cchar_t *clw=(cchar_t*)alloca(ll*sizeof(cchar_t));
+   cchar_t *clwp;
+#endif
 
    if(hex)
    {
@@ -459,47 +561,116 @@ void  Redisplay(num line,offs ptr,num limit)
       {
          lptr=ptr;
 
-         clp=cl;
+	 if(!mb_mode)
+	 {
+	    clp=cl;
 
-	 if(!EofAt(ptr) || line==(Size()-ScreenTop.Offset()+15)/16)
-         {
-            sprintf(s,"%08lX   ",(unsigned long)ptr);
-	    for(sp=s; *sp; sp++)
-	       *clp++=norm_attr->n_attr|*sp;
-         }
-         for(i=0; i<16 && !EofAt(ptr); i++)
-         {
-            if(EofAt(ptr))
-               break;
-            sprintf(s,"%02X",CharAt(ptr));
-            ca=(InBlock(ptr)?blk_attr:norm_attr);
-            if(ascii && ptr==Offset() && ShowMatchPos)
-               ca=SHADOW_ATTR;
-	    *clp++=ca->n_attr|s[0];
-	    *clp++=ca->n_attr|s[1];
-            ptr++;
-	    char b=' '; // ((ptr&15)!=8 ? ' ' : '-');
-	    if(InBlock(ptr-1) && InBlock(ptr) && (ptr&15))
-               *clp++=blk_attr->n_attr|b;
-            else
-               *clp++=norm_attr->n_attr|b;
-         }
-	 while(clp-cl<AsciiPos)
-	    *clp++=norm_attr->n_attr|' ';
-         ptr=lptr;
-         for(i=0; i<16 && !EofAt(ptr); i++,ptr++)
-         {
-            ca=(InBlock(ptr)?blk_attr:norm_attr);
-            if(!ascii && ptr==Offset() && ShowMatchPos)
-               ca=SHADOW_ATTR;
-	    *clp++=visualize(ca,CharAt_NoCheck(ptr)|ca->n_attr);
-         }
-	 // clear the rest of line
-	 for(i=TextWinWidth-(clp-cl); i>0; i--)
-            *clp++=norm_attr->n_attr|' ';
+	    if(!EofAt(ptr) || line==(Size()-ScreenTop.Offset()+15)/16)
+	    {
+	       sprintf(s,"%08lX   ",(unsigned long)ptr);
+	       for(sp=s; *sp; sp++)
+		  *clp++=norm_attr->n_attr|*sp;
+	    }
+	    for(i=0; i<16 && !EofAt(ptr); i++)
+	    {
+	       if(EofAt(ptr))
+		  break;
+	       sprintf(s,"%02X",CharAt(ptr));
+	       ca=(InBlock(ptr)?blk_attr:norm_attr);
+	       if(ascii && ptr==Offset() && ShowMatchPos)
+		  ca=SHADOW_ATTR;
+	       *clp++=ca->n_attr|s[0];
+	       *clp++=ca->n_attr|s[1];
+	       ptr++;
+	       char b=' '; // ((ptr&15)!=8 ? ' ' : '-');
+	       if(InBlock(ptr-1) && InBlock(ptr) && (ptr&15))
+		  *clp++=blk_attr->n_attr|b;
+	       else
+		  *clp++=norm_attr->n_attr|b;
+	    }
+	    while(clp-cl<AsciiPos)
+	       *clp++=norm_attr->n_attr|' ';
+	    ptr=lptr;
+	    for(i=0; i<16 && !EofAt(ptr); i++,ptr++)
+	    {
+	       ca=(InBlock(ptr)?blk_attr:norm_attr);
+	       if(!ascii && ptr==Offset() && ShowMatchPos)
+		  ca=SHADOW_ATTR;
+	       *clp++=visualize(ca,CharAt_NoCheck(ptr)|ca->n_attr);
+	    }
+	    // clear the rest of line
+	    for(i=TextWinWidth-(clp-cl); i>0; i--)
+	       *clp++=norm_attr->n_attr|' ';
 
-	 attrset(0);
-	 mvaddchnstr(TextWinY+line,TextWinX,cl,TextWinWidth);
+	    attrset(0);
+	    mvaddchnstr(TextWinY+line,TextWinX,cl,TextWinWidth);
+	 }
+#ifdef USE_MULTIBYTE_CHARS
+	 else // mb_mode
+	 {
+	    memset(clw,0,ll*sizeof(cchar_t));
+	    clwp=clw;
+	    if(!EofAt(ptr) || line==(Size()-ScreenTop.Offset()+15)/16)
+	    {
+	       sprintf(s,"%08lX   ",(unsigned long)ptr);
+	       for(sp=s; *sp; sp++)
+	       {
+		  clwp->attr=norm_attr->n_attr;
+		  clwp->chars[0]=*sp;
+		  clwp++;
+	       }
+	    }
+	    for(i=0; i<16 && !EofAt(ptr); i++)
+	    {
+	       if(EofAt(ptr))
+		  break;
+	       sprintf(s,"%02X",CharAt(ptr));
+	       ca=(InBlock(ptr)?blk_attr:norm_attr);
+	       if(ascii && ptr==Offset() && ShowMatchPos)
+		  ca=SHADOW_ATTR;
+	       clwp->attr=ca->n_attr;
+	       clwp->chars[0]=s[0];
+	       clwp++;
+	       clwp->attr=ca->n_attr;
+	       clwp->chars[0]=s[1];
+	       clwp++;
+	       ptr++;
+	       char b=' '; // ((ptr&15)!=8 ? ' ' : '-');
+	       ca=((InBlock(ptr-1) && InBlock(ptr) && (ptr&15))?blk_attr:norm_attr);
+	       clwp->attr=ca->n_attr;
+	       clwp->chars[0]=b;
+	       clwp++;
+	    }
+	    while(clwp-clw<AsciiPos)
+	    {
+	       clwp->attr=norm_attr->n_attr;
+	       clwp->chars[0]=' ';
+	       clwp++;
+	    }
+	    ptr=lptr;
+	    for(i=0; i<16 && !EofAt(ptr); i++,ptr++)
+	    {
+	       ca=(InBlock(ptr)?blk_attr:norm_attr);
+	       if(!ascii && ptr==Offset() && ShowMatchPos)
+		  ca=SHADOW_ATTR;
+	       wchar_t ch=CharAt_NoCheck(ptr);
+	       clwp->attr=ca->n_attr;
+	       clwp->chars[0]=visualize_wchar(ch);
+	       if(clwp->chars[0]!=ch)
+		  clwp->attr=ca->so_attr;
+	       clwp++;
+	    }
+	    // clear the rest of line
+	    for(i=TextWinWidth-(clwp-clw); i>0; i--)
+	    {
+	       clwp->attr=norm_attr->n_attr;
+	       clwp->chars[0]=' ';
+	       clwp++;
+	    }
+	    attrset(0);
+	    mvadd_wchnstr(TextWinY+line,TextWinX,clw,TextWinWidth);
+	 }
+#endif // USE_MULTIBYTE_CHARS
       }
    }
    else /* !hex */
@@ -595,56 +766,143 @@ void  Redisplay(num line,offs ptr,num limit)
 	 else
 	    col=n.Col()-ScrShift;
 
-	 clp=cl;
+	 if(!mb_mode)
+	 {
+	    clp=cl;
+	    for( ; col<TextWinWidth && !EolAt(ptr); ptr++)
+	    {
+	       if(col>-TabSize)
+		  ca=FindPosAttr(ptr,line,col,hlp);
 
-         for( ; col<TextWinWidth && !EolAt(ptr); ptr++)
-         {
-	    if(col>-TabSize)
-	       ca=FindPosAttr(ptr,line,col,hlp);
-
-	    byte ch=CharAt_NoCheck(ptr);
-            if(ch=='\t')
-            {
-               i=Tabulate(col+ScrShift)-ScrShift;
-               if(i>0)
+	       byte ch=CharAt_NoCheck(ptr);
+	       if(ch=='\t')
 	       {
-		  if(col<0)
-		     col=0;
-		  *clp++ = ca->n_attr|' ';
-		  col++;
-		  while(col<i && col<TextWinWidth)
+		  i=Tabulate(col+ScrShift)-ScrShift;
+		  if(i>0)
 		  {
-		     ca=FindPosAttr(ptr,line,col,hlp);
+		     if(col<0)
+			col=0;
 		     *clp++ = ca->n_attr|' ';
 		     col++;
+		     while(col<i && col<TextWinWidth)
+		     {
+			ca=FindPosAttr(ptr,line,col,hlp);
+			*clp++ = ca->n_attr|' ';
+			col++;
+		     }
 		  }
+		  else
+		     col=i;
 	       }
 	       else
-		  col=i;
-            }
-            else
-            {
-               if(col>=0)
-		  *clp++=visualize(ca,ch|ca->n_attr);
-	       col++;
-            }
-	    if(hlp)
-	       hlp++;
-         }
-	 if(col<0)
-	    col=0;
+	       {
+		  if(col>=0)
+		     *clp++=visualize(ca,ch|ca->n_attr);
+		  col++;
+	       }
+	       if(hlp)
+		  hlp++;
+	    }
+	    if(col<0)
+	       col=0;
 
-	 if(EofAt(ptr))
-	    hlp=0;
+	    if(EofAt(ptr))
+	       hlp=0;
 
-         for( ; col<TextWinWidth; col++)
-	 {
-	    ca=FindPosAttr(ptr,line,col,hlp);
-            *clp++ = ca->n_attr|' ';
+	    for( ; col<TextWinWidth; col++)
+	    {
+	       ca=FindPosAttr(ptr,line,col,hlp);
+	       *clp++ = ca->n_attr|' ';
+	    }
+
+	    attrset(0);
+	    mvaddchnstr(TextWinY+line,TextWinX,cl,TextWinWidth);
 	 }
+#ifdef USE_MULTIBYTE_CHARS
+	 else // mb_mode
+	 {
+	    memset(clw,0,ll*sizeof(cchar_t));
+	    clwp=clw;
+	    for( ; col<TextWinWidth && !EolAt(ptr); ptr+=MBCharSize)
+	    {
+	       wchar_t ch=WCharAt(ptr);
 
-	 attrset(0);
-	 mvaddchnstr(TextWinY+line,TextWinX,cl,TextWinWidth);
+	       if(col>-TabSize)
+		  ca=FindPosAttr(ptr,line,col,hlp);
+
+	       if(ch=='\t')
+	       {
+		  i=Tabulate(col+ScrShift)-ScrShift;
+		  if(i>0)
+		  {
+		     if(col<0)
+			col=0;
+		     clwp->attr=ca->n_attr;
+		     clwp->chars[0]=' ';
+		     clwp++;
+		     col++;
+		     while(col<i && col<TextWinWidth)
+		     {
+			ca=FindPosAttr(ptr,line,col,hlp);
+			clwp->attr=ca->n_attr;
+			clwp->chars[0]=' ';
+			clwp++;
+			col++;
+		     }
+		  }
+		  else
+		     col=i;
+	       }
+	       else
+	       {
+		  if(col>=0)
+		  {
+		     if(MBCharWidth==0 && clwp>clw)
+		     {
+			int a=0;
+			while(clwp[-1].chars[a] && a<CCHARW_MAX)
+			   a++;
+			if(a<CCHARW_MAX)
+			   clwp[-1].chars[a]=ch;
+		     }
+		     else if(MBCharWidth>0)
+		     {
+			clwp->attr=ca->n_attr;
+			clwp->chars[0]=visualize_wchar(ch);
+			if(clwp->chars[0]!=ch)
+			   clwp->attr=ca->so_attr;
+			else if(MBCharSize==1 && !chset_isprint(ch))
+			{
+			   chtype v=visualize(ca,CharAt(ptr)|ca->n_attr);
+			   clwp->chars[0]=v&A_CHARTEXT;
+			   clwp->attr=v&~A_CHARTEXT;
+			}
+			clwp++;
+		     }
+		  }
+		  col+=MBCharWidth;
+	       }
+	       if(hlp)
+		  hlp+=MBCharSize;
+	    }
+	    if(col<0)
+	       col=0;
+
+	    if(EofAt(ptr))
+	       hlp=0;
+
+	    for( ; col<TextWinWidth; col++)
+	    {
+	       ca=FindPosAttr(ptr,line,col,hlp);
+	       clwp->attr=ca->n_attr;
+	       clwp->chars[0]=' ';
+	       clwp++;
+	    }
+
+	    attrset(0);
+	    mvadd_wchnstr(TextWinY+line,TextWinX,clw,TextWinWidth);
+	 } // end of mb_mode
+#endif
 
 	 if(++line>=limit)
 	    break;
