@@ -17,7 +17,7 @@
  */
 
 /* Search & replace for the editor */
-/* $Id: search.cc,v 1.36 2006/11/29 07:36:35 lav Exp $ */
+/* $Id: search.cc,v 1.37 2007/08/08 05:54:53 lav Exp $ */
 
 #include <config.h>
 #include <memory.h>
@@ -371,19 +371,15 @@ search_again:
    else
       fndlen=regs.end[0]-res;
    CurrentPos=fndind;
-
-   if(word_bounds)
+   if(fndlen==0
+   || word_bounds && (isalnum(CharRel(-1)) || isalnum(CharRel(fndlen))))
    {
-      if(isalnum(CharRel(-1)) || isalnum(CharRel(fndlen)))
-      {
-	 if(dir==FORWARD)
-	    srchpos=CurrentPos+1;
-	 else // dir==BACKWARD
-	    srchpos=CurrentPos-1;
-	 goto search_again;
-      }
+      if(dir==FORWARD)
+	 srchpos=CurrentPos+1;
+      else // dir==BACKWARD
+	 srchpos=CurrentPos-1;
+      goto search_again;
    }
-
    return(TRUE);
 }
 
