@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2000 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1993-2006 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: user.cc,v 1.34 2001/05/22 10:58:24 lav Exp $ */
+/* $Id: user.cc,v 1.55 2006/11/29 07:10:25 lav Exp $ */
 
 #include <config.h>
 #include <ctype.h>
@@ -36,6 +36,7 @@
 #include "format.h"
 #include "about.h"
 #include "bm.h"
+#include "undo.h"
 
 void  UserDeleteToEol()
 {
@@ -127,7 +128,7 @@ void  UserCharRight()
    if(hex)
    {
       if(ascii)
-         MoveRight();
+	 MoveRight();
       else
       {
          if(right)
@@ -178,12 +179,12 @@ void  UserCopyFromDown()
       }
       if(tp.Col()==oc && !EolAt(tp.Offset()))
       {
-         char  ch=CharAt(tp.Offset());
+         wchar_t ch=WCharAt(tp.Offset());
          PreUserEdit();
          if(insert)
-            InsertChar(ch);
+            InsertWChar(ch);
          else
-            ReplaceCharExtMove(ch);
+            ReplaceWCharExtMove(ch);
          flag|=REDISPLAY_LINE;
          return;
       }
@@ -212,12 +213,12 @@ void  UserCopyFromUp()
       }
       if(tp.Col()==oc && !EolAt(tp.Offset()))
       {
-         char  ch=CharAt(tp.Offset());
+         wchar_t ch=WCharAt(tp.Offset());
          PreUserEdit();
          if(insert)
-            InsertChar(ch);
+            InsertWChar(ch);
          else
-            ReplaceCharExtMove(ch);
+            ReplaceWCharExtMove(ch);
          flag|=REDISPLAY_LINE;
          return;
       }
@@ -273,8 +274,7 @@ void  UserBackwardDeleteWord()
 {
    if(View || hex)
       return;
-   if(!isalnum(CharRel(-1)) && !isrussian(CharRel(-1))
-   && CharRel(-1)!=' ' && CharRel(-1)!='\t')
+   if(!IsAlNumLeft() && CharRel(-1)!=' ' && CharRel(-1)!='\t')
    {
       UserBackSpace();
    }
@@ -284,12 +284,12 @@ void  UserBackwardDeleteWord()
       if(!Bol() && (CharRel(-1)==' ' || CharRel(-1)=='\t'))
       {
 	 while(!Bol() && (CharRel(-1)==' ' || CharRel(-1)=='\t'))
-	    BackSpace();
+	    DeleteBlock(1,0);
       }
       else
       {
-	 while(!Bol() && (isalnum(CharRel(-1)) || isrussian(CharRel(-1))))
-            BackSpace();
+	 while(!Bol() && IsAlNumLeft())
+	    DeleteBlock(MBCharSize,0);
       }
    }
    stdcol=GetCol();
@@ -300,8 +300,7 @@ void  UserForwardDeleteWord()
 {
    if(View || hex)
       return;
-   if(!isalnum(Char()) && !isrussian(Char())
-   && Char()!=' ' && Char()!='\t')
+   if(!IsAlNumRel(0) && Char()!=' ' && Char()!='\t')
       UserDeleteChar();
    else
    {
@@ -309,12 +308,12 @@ void  UserForwardDeleteWord()
       if(!Eol() && (Char()==' ' || Char()=='\t'))
       {
 	 while(!Eol() && (Char()==' ' || Char()=='\t'))
-	    DeleteChar();
+	    DeleteBlock(0,1);
       }
       else
       {
-	 while(!Eol() && (isalnum(Char()) || isrussian(Char())))
-            DeleteChar();
+	 while(!Eol() && IsAlNumRel(0))
+            DeleteBlock(0,MBCharSize);
       }
    }
    stdcol=GetCol();
@@ -325,15 +324,15 @@ void  UserDeleteWord()
 {
    if(View || hex)
       return;
-   if(!isalnum(Char()) && !isrussian(Char()))
+   if(!IsAlNumRel(0))
       UserForwardDeleteWord();
    else
    {
       PreUserEdit();
-      while(!Eol() && (isalnum(Char()) || isrussian(Char())))
-         DeleteChar();
-      while(!Bol() && (isalnum(CharRel(-1)) || isrussian(CharRel(-1))))
-         BackSpace();
+      while(!Eol() && IsAlNumRel(0))
+         DeleteBlock(0,MBCharSize);
+      while(!Bol() && IsAlNumLeft())
+	 DeleteBlock(MBCharSize,0);
    }
    stdcol=GetCol();
    flag|=REDISPLAY_LINE;
@@ -346,12 +345,18 @@ void  UserMarkWord()
 
    offs word_begin=CurrentPos;
    offs word_end=CurrentPos;
-   while(!BofAt(word_begin) && isalnum(CharAt(word_begin-1)))
-      word_begin--;
-   while(!EofAt(word_end) && isalnum(CharAt(word_end)))
-      word_end++;
+   while(!BofAt(word_begin))
+   {
+      MBCheckLeftAt(word_begin);
+      if(IsAlNumAt(word_begin-MBCharSize))
+	 word_begin--;
+      else
+	 break;
+   }
+   while(!EofAt(word_end) && IsAlNumAt(word_end))
+      word_end+=MBCharSize;
    if(word_end==word_begin)
-      word_end++;
+      word_end+=MBCharSize;
    BlockBegin=word_begin;
    BlockEnd=word_end;
    hide=FALSE;
@@ -522,10 +527,10 @@ void  UserWordLeft()
       MoveLeft();
    else
    {
-      while(!Bof() && !isalnum(CharRel(-1)) && !isrussian(CharRel(-1)))
-         MoveLeft();
-      while(!Bof() && (isalnum(CharRel(-1)) || isrussian(CharRel(-1))))
-         MoveLeft();
+      while(!Bof() && !IsAlNumLeft())
+         CurrentPos-=MBCharSize;
+      while(!Bof() && IsAlNumLeft())
+         CurrentPos-=MBCharSize;
    }
    stdcol=GetCol();
 }
@@ -535,10 +540,10 @@ void  UserWordRight()
       MoveRight();
    else
    {
-      while(!Eof() && !isalnum(Char()) && !isrussian(Char()))
-         MoveRight();
-      while(!Eof() && (isalnum(Char()) || isrussian(Char())))
-         MoveRight();
+      while(!Eof() && !IsAlNumRel(0))
+         CurrentPos+=MBCharSize;
+      while(!Eof() && IsAlNumRel(0))
+         CurrentPos+=MBCharSize;
    }
    stdcol=GetCol();
 }
@@ -557,7 +562,10 @@ void  UserCommentLine()
       return;
 
    ToLineBegin();
-   if(Suffix(FileName,".cc"))
+   if(Suffix(FileName,".cc")
+   || Suffix(FileName,".cpp")
+   || Suffix(FileName,".cxx")
+   || Suffix(FileName,".java"))
    {
       if(Char()=='/' && CharRel(1)=='/')
       {
@@ -570,8 +578,28 @@ void  UserCommentLine()
 	 InsertBlock("// ",3);
       }
    }
+   else if(Suffix(FileName,".sql"))
+   {
+      if(Char()=='-' && CharRel(1)=='-')
+      {
+	 DeleteBlock(0,2);
+	 if(Char()==' ')
+	    DeleteBlock(0,1);
+      }
+      else
+      {
+	 InsertBlock("-- ",3);
+      }
+   }
    else if(Suffix(FileName,".c") || Suffix(FileName,".h"))
    {
+      if(Char()=='/' && CharRel(1)=='/')
+      {
+	 DeleteBlock(0,2);
+	 if(Char()==' ')
+	    DeleteBlock(0,1);
+	 goto done;
+      }
       if(Char()=='/' && CharRel(1)=='*')
       {
 	 unc=1;
@@ -590,6 +618,20 @@ void  UserCommentLine()
 	 InsertBlock("/*",2);
       }
    }
+   else // default
+   {
+      if(Char()=='#')
+      {
+	 DeleteChar();
+	 if(Char()==' ')
+	    DeleteChar();
+      }
+      else
+      {
+	 InsertBlock("# ",2);
+      }
+   }
+done:
    CurrentPos=op;
    stdcol=GetCol();
    flag|=REDISPLAY_LINE;
@@ -674,6 +716,12 @@ void  UserLineEnd()
 {
    ToLineEnd();
    stdcol=GetCol();
+   if(autoindent && Text && Bol() && !Bof())
+   {
+      bool old_modified=modified;
+      InsertAutoindent(TextPoint(CurrentPos-EolSize).Col());
+      modified=old_modified;
+   }
 }
 void  UserFileBegin()
 {
@@ -909,6 +957,7 @@ int   file_check(char *fn)
 void    UserLoad()
 {
    char  newname[256];
+   newname[0]=0;
 
    if(getstring("Load: ",newname,sizeof(newname)-1,&LoadHistory)>0)
    {
@@ -935,6 +984,7 @@ void    UserLoad()
 int   UserSaveAs()
 {
    char  newname[256];
+   newname[0]=0;
 
    if(getstring("Save as: ",newname,sizeof(newname)-1,&LoadHistory,NULL,NULL)>0)
    {
@@ -951,20 +1001,18 @@ int   UserSaveAs()
 }
 void  UserSwitch()
 {
-   char  newname[256];
-   HistoryLine *prev;
-
    LoadHistory.Open();
    LoadHistory.Prev();
-   LoadHistory.Prev();
-   prev=LoadHistory.Curr();
+   const HistoryLine *prev=LoadHistory.Prev();
    if(prev==NULL)
    {
       UserLoad();
       return;
    }
 
-   strcpy(newname,prev->line);
+   char newname[256];
+   strncpy(newname,prev->get_line(),255);
+   newname[255]=0;
 
    if(ChooseFileName(newname)<0)
       return;
@@ -1122,7 +1170,8 @@ void  UserNewLine()
 {
    if(View)
       return;
-   if(autoindent)
+
+   if(autoindent && !CheckPending())
       UserAutoindent();
    else
    {
@@ -1134,75 +1183,33 @@ void  UserNewLine()
 
 void  UserAutoindent()
 {
-   int   UseTabsNow=UseTabs;
-   offs  ptr;
-   num   cnt;
-   num   oldcol;
-   num   oldmargin;
-   num   newmargin=0;
-
    if(View)
       return;
 
-   oldcol=GetCol();
+   num oldcol=GetCol();
    if(Text && Eol() && oldcol<stdcol)
       oldcol=stdcol;
 
-   offs o=Offset();
-   for(;;)
-   {
-      oldmargin=MarginSizeAt(o);
-      if(BofAt(o) || oldmargin!=-1)
-	 break;
-      o=PrevLine(o);
-   }
-   if(TabsInMargin)
-      UseTabsNow=1;
+   bool do_indent=true;
 
-   if(oldmargin==oldcol)
-      newmargin=oldmargin;
-   else if(oldcol==0)
-      newmargin=0;
-   else if(oldmargin==-1)
-   {
+   if(MarginSizeAt(Offset())==-1)
       DeleteBlock(Offset()-LineBegin(Offset()),0);
-      newmargin=oldcol/IndentSize*IndentSize;
-   }
    else
    {
-      for(ptr=Offset(); !EolAt(ptr) && (CharAt(ptr)==' ' || CharAt(ptr)=='\t'); ptr++);
+      offs ptr;
+      for(ptr=Offset(); !EolAt(ptr) && (CharAt(ptr)==' ' || CharAt(ptr)=='\t'); ptr++)
+	 ;
       if(EolAt(ptr))
-      {
          DeleteToEOL();
-         newmargin=oldmargin;
-      }
       else
-         newmargin=0;
+         do_indent=false;
    }
 
    NewLine();
-   flag|=REDISPLAY_AFTER;
-
-   cnt=newmargin;
-   if(Text && Eol() && !(!UseTabs && UseTabsNow))
-   {
-      stdcol=cnt;
-      return;
-   }
-   if(UseTabsNow)
-   {
-       while(cnt>=TabSize)
-       {
-           cnt-=TabSize;
-           InsertChar('\t');
-       }
-   }
-   while(cnt>0)
-   {
-       cnt--;
-       InsertChar(' ');
-   }
    stdcol=GetCol();
+   flag|=REDISPLAY_AFTER;
+   if(do_indent)
+      InsertAutoindent(oldcol);
 }
 
 void  UserUndelete()
@@ -1213,6 +1220,41 @@ void  UserUndelete()
    flag=REDISPLAY_ALL;
    stdcol=GetCol();
 }
+void  UserUndo()
+{
+   if(View)
+      return;
+   if(!undo.Enabled())
+   {
+      UserUndelete();
+      return;
+   }
+   undo.UndoGroup();
+   flag=REDISPLAY_ALL;
+}
+void  UserRedo()
+{
+   if(View)
+      return;
+   undo.RedoGroup();
+   flag=REDISPLAY_ALL;
+   stdcol=GetCol();
+}
+void  UserUndoStep()
+{
+   if(View)
+      return;
+   undo.UndoOne();
+   flag=REDISPLAY_ALL;
+}
+void  UserRedoStep()
+{
+   if(View)
+      return;
+   undo.RedoOne();
+   flag=REDISPLAY_ALL;
+   stdcol=GetCol();
+}
 
 void  UserInsertChar(char ch)
 {
@@ -1230,6 +1272,31 @@ void  UserInsertChar(char ch)
       flag|=REDISPLAY_LINE;
    stdcol=GetCol();
 }
+void  UserReplaceChar(char ch)
+{
+   if(View)
+      return;
+   PreUserEdit();
+
+   if(!hex && Eol())
+      flag|=REDISPLAY_AFTER;
+
+   if(buffer_mmapped || hex || !mb_mode)
+      ReplaceCharMove(ch);
+   else
+   {
+      InsertChar(ch);
+      MBCheckLeft();
+      if(!MBCharInvalid)
+	 DeleteChar();
+   }
+
+   if(!hex && Bol())
+      flag|=REDISPLAY_AFTER;
+   else
+      flag|=REDISPLAY_LINE;
+   stdcol=GetCol();
+}
 
 void  UserInsertControlChar(char ch)
 {
@@ -1256,6 +1323,19 @@ void  UserInsertControlChar(char ch)
    }
    stdcol=GetCol();
 }
+void  UserInsertString(const char *s,int len)
+{
+   while(len-->0)
+      UserInsertControlChar(*s++);
+}
+void UserInsertWChar(wchar_t ch)
+{
+   char buf[MB_CUR_MAX+1];
+   int len=wctomb(buf,ch);
+   if(len<=0)
+      return;
+   UserInsertString(buf,len);
+}
 
 void  UserEnterControlChar()
 {
@@ -1301,18 +1381,46 @@ void  UserRefreshScreen()
 }
 
 void  UserChooseChar()
+{
+   if(mb_mode && !hex)
+      UserChooseWChar();
+   else
+      UserChooseByte();
+}
+void  UserChooseByte()
 {
    int   ch=choose_ch();
    if(ch!=-1)
       UserInsertControlChar(ch);
 }
+void  UserChooseWChar()
+{
+   wchar_t ch=choose_wch();
+   if(ch!=-1)
+      UserInsertWChar(ch);
+}
 
 void  UserInsertCharCode()
 {
-   int   ch=getcode();
+   if(mb_mode && !hex)
+      UserInsertWCharCode();
+   else
+      UserInsertByteCode();
+}
+void  UserInsertByteCode()
+{
+   if(View)
+      return;
+   int ch=getcode_char();
    if(ch!=-1)
       UserInsertControlChar(ch);
 }
+void  UserInsertWCharCode()
+{
+   wchar_t ch=getcode_wchar();
+   if(ch!=-1)
+      UserInsertWChar(ch);
+}
 
 static int base_editmode=-1;
 
@@ -1402,11 +1510,14 @@ void  UserBlockPrefixIndent()
    flag=REDISPLAY_ALL;
 }
 
+History	 ShellHistory;
+History	 PipeHistory;
+
 void  UserShellCommand()
 {
    static char str[256];
 
-   if(getstring("Shell-Command: ",str,sizeof(str)-1)<1)
+   if(getstring("Shell-Command: ",str,sizeof(str)-1,&ShellHistory)<1)
       return;
 
    cmd(str,0,1);
@@ -1423,7 +1534,7 @@ void  UserPipeBlock()
    if(hide || rblock || View)
       return;
 
-   if(getstring("Pipe through: ",filter,sizeof(filter)-1,NULL,NULL,NULL)<1)
+   if(getstring("Pipe through: ",filter,sizeof(filter)-1,&PipeHistory,NULL,NULL)<1)
       return;
 
    MessageSync("Piping...");
@@ -1592,7 +1703,10 @@ void UserOptimizeText()
       {
          CurrentPos=ptr;
          while(!Bol() && (CharRel_NoCheck(-1)==' ' || CharRel_NoCheck(-1)=='\t'))
+	 {
             BackSpace();
+	    ptr--;
+	 }
       }
    }
    CurrentPos=TextEnd;
