@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-1997 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1993-2005 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,12 +19,24 @@
 #ifndef WINDOW_H
 #define WINDOW_H
 
+#ifdef USE_MULTIBYTE_CHARS
+#define win_cell cchar_t
+#define win_cell_set_attrs(ch,a) (ch)->attr=((ch)->attr&A_ALTCHARSET)|((a)&~(A_CHARTEXT|A_ALTCHARSET))
+#define scr_get_cell(y,x,out) mvin_wch(y,x,out)
+#define scr_put_cell(y,x,ch)  mvadd_wchnstr(y,x,ch,1)
+#else
+#define win_cell chtype
+#define win_cell_set_attrs(ch,a) (*ch)=((*ch)&(A_CHARTEXT|A_ALTCHARSET))|((a)&~(A_CHARTEXT|A_ALTCHARSET))
+#define scr_get_cell(y,x,out) (*out)=mvinch(y,x)
+#define scr_put_cell(y,x,ch)  mvaddch(y,x,*(ch))
+#endif
+
 typedef struct  win
 {
-    chtype  *buf;
+    win_cell *buf;
     int     x,y;
     int     w,h;
-    struct attr *a;
+    const attr *a;
     const char  *title;
     struct win  *prev;
     int     flags;
@@ -38,7 +50,7 @@ typedef struct  win
 /* window flags */
 #define NOSHADOW    1
 
-WIN   *CreateWin(int x,int y,unsigned w,unsigned h,struct attr *a,
+WIN   *CreateWin(int x,int y,unsigned w,unsigned h,const attr *a,
                  const char *title,int flags=0);
 void  DisplayWin(WIN *);
 void  CloseWin();
@@ -49,11 +61,19 @@ void  GotoXY(int x,int y);
 void  Clear();
 void  PutStr(int x,int y,const char *s);
 void  PutCh(int x,int y,chtype ch);
+#ifdef USE_MULTIBYTE_CHARS
+void  PutWCh(int x,int y,wchar_t ch);
+void  PutCCh(int x,int y,cchar_t *ch);
+#define PutACS(x,y,a) do { if(mb_mode) PutCCh(x,y,WACS_##a); else PutCh(x,y,ACS_##a); } while(0)
+#else
+# define PutWCh(x,y,c) PutCh(x,y,c)
+#define PutACS(x,y,a) PutCh(x,y,ACS_##a)
+#endif
 
-extern struct attr *curr_attr;
+extern const struct attr *curr_attr;
 extern WIN *Upper;
 
-inline void  SetAttr(struct attr *a)
+static inline void SetAttr(const struct attr *a)
 {
    curr_attr=a;
    attrset(a->n_attr);
